# -----------------------------------------
# * Env *
# -----------------------------------------
import os

MOKO_PATH = os.path.abspath(os.path.dirname(__file__))

# -----------------------------------------
# * NLP *
# -----------------------------------------

# -----------------------------------------
# moko.nlp.processing
# -----------------------------------------
STOPWORD_PATH = MOKO_PATH + "/data/stopword.txt"
ADVERB_PATH = MOKO_PATH + "/data/adverb.txt"
SYNONYM_PATH = MOKO_PATH + "/data/synonym.csv"
CHARACTER_PATH = MOKO_PATH + "/data/char_synonym.txt"
CONCEPT_NOUN_PATH = MOKO_PATH + "/data/entities.csv"
USER_DICT_PATH = MOKO_PATH + "/data/user_dict.txt"

SPACING_MODEL_PATH = MOKO_PATH + "/model/soyspacing_model"
SPACING_RULE_PATH = MOKO_PATH + "/data/editorial_spacing_rules.txt"

NUM_CHAR_LIST = ['一', '壹', '二', '貳', '三', '四', '五', '六', '七', '八', '九', '十', '百', '千', '萬']
PREFIX_LIST = ['其', '第']
COUNT_NOUN_LIST = ['月', '日', '百', '千', '萬', '時', '分', '人', '曰', '次', '種']
DEP_NOUN_LIST = ['等']
LAST_CHAR_LIST = ['曰']
EXCEPTION_WORD_LIST = ['平等', '少年', '靑年','中年','凶年', ]
DEMON_WORD_LIST = ['其','他','彼','此','之']

MODEL_CONFIG = {
    "hidden_size": 768,
    "dropout": 0.1,
}

VOCAB = {
    "padding_token": "[PAD]",
    "cls_token": "[CLS]",
    "sep_token": "[SEP]",
    "unk_token": "[UNK]"
}



import unicodedata
import re
import hanja

import moko.configure as con

cp_path = con.CHARACTER_PATH
num_char = con.NUM_CHAR_LIST
prefix = con.PREFIX_LIST
exp_noun = con.EXCEPTION_WORD_LIST
counter_noun = con.COUNT_NOUN_LIST
last_char = con.LAST_CHAR_LIST
demon_noun = con.DEMON_WORD_LIST

def pre_processing(text):
    text = re.sub('[▣◎△◆△◆.,「」『』【】{}〔〕(/)▲◇▼●○"]', ' ', text)
    text = re.sub('[、]', ' ', text)
    text = text.strip().replace('\n', ' ')
    return text

def load_dict():
    character_pair = {}
    with open(cp_path, 'r', encoding = 'utf-8') as character_file:
        while True:
            line = character_file.readline().replace('\n', '')
            if not line: break
            pair_list = line.split('\t')
            old_character = pair_list[0]
            new_character = pair_list[1]
            character_pair[old_character] = new_character
    return character_pair

def unicode_normalize(text, character_pair={}):
    for char in text:
        n_char = unicodedata.normalize('NFC',char) 
        text = text.replace(char,n_char)
        if char in character_pair.keys():
            text = text.replace(char, character_pair[char])
    return text

def num_word_check(word):
    cnt = 0
    word_length = len(word)
    is_num_word, is_skip = False, False
    for char in word:
        if char in num_char : cnt += 1
    if word_length == cnt : is_num_word = True

    for c in prefix:
        if word_length > 1 and word.startswith(c) and word_length-1 == cnt: is_skip = True 

    for c in counter_noun:
        if word_length > 1 and word.endswith(c) and word_length-1 == cnt: is_skip = True
    
    return is_num_word, is_skip, cnt

def prefix_remove(word):
    cnt = 1
    is_measure, is_num = False, False
    for s in prefix:
        if word.startswith(s): is_measure = True
    
    for char in word[1:]:
        if is_measure and char in num_char: 
            is_num = True
            cnt+=1
        else:
            is_num = False
            if is_measure and is_num==False: word = word[cnt-1:]
            continue
    return word

def date_check(word, cnt):
    word_length = len(word)
    is_date, is_year = False, False
    regex_lst = ['\w{2,4}年\w{1,3}月\w{1,3}日$', '\w{1,3}月\w{1,3}日$', '\w{1,4}年\w{1,3}月$']
    for regex in regex_lst:
        r = re.compile(regex)
        m = r.match(word)
        if m: is_date = True
    for char in last_char: # exception word
        if word_length > 1 and word.endswith(char) and (word_length-1 == cnt) and word not in exp_noun : is_year = True
    return is_date, is_year

def last_char_remove(word):
    word_length = len(word)
    for char in last_char: # exception word
        if word_length > 2 and (word.endswith(char) and word not in exp_noun): word = word[:-1]
    return word

def stopword_remove(word_lst, stopword_lst):
    lst = []
    for w in word_lst:
        if w in stopword_lst: continue
        else: lst.append(w)
    return lst

def adverb_remove(word_lst, adverb_lst):
    lst = []
    for w in word_lst:
        translated = hanja.translate(w,'substitution')
        if translated in adverb_lst: continue
        else: lst.append(w)
    return lst



from hanja import hangul
from hanja import split_hanja

from soyspacing.countbase import RuleDict
from soyspacing.countbase import CountSpace

import moko.processor as processor
import moko.configure as con

#from moko.dict import stopword_remove,adverb_remove
from moko.utils_io import list_reader

adv_path = con.ADVERB_PATH
stw_path = con.STOPWORD_PATH
cpt_path = con.CONCEPT_NOUN_PATH
usr_path = con.USER_DICT_PATH

concept_lst = list_reader(cpt_path)
adverb_lst = list_reader(adv_path, False)
stopword_lst = list_reader(stw_path, False)
usrword_lst = list_reader(usr_path, False)

model = CountSpace()
model.load_model(con.SPACING_MODEL_PATH, json_format=False)
rule_dict = RuleDict(con.SPACING_RULE_PATH)

def noun_chunk_dict(text, stopword_lst=stopword_lst, usrword_lst=usrword_lst, char_num = 4):
    lst = []
    uni = processor.load_dict()

    line = processor.pre_processing(text)
    line = processor.unicode_normalize(line,uni)

    word_list = hanja_text_to_list(line)
    for word in word_list:
        word_length = len(word)
        word = word.replace('\n', '').replace(' ', '').replace(',','')
        
        is_all_hanja = True
        for char in word:
            if hangul.is_hangul(char): is_all_hanja = False
            
        is_num_word, is_skip, cnt = processor.num_word_check(word)
        if is_num_word or is_skip: continue
        is_date, is_year = processor.date_check(word, cnt)
        if is_date or is_year: continue
        
        word = processor.last_char_remove(word)
        word = processor.prefix_remove(word)
        
        if is_all_hanja and word_length > 1 and word.isalpha()== True: 
            if word_length > char_num: # long word
                '''
                sent_corrected, tags = model.correct(word, rules=rule_dict) # soyspacing 모델, space_rule을 적용
                candidate = [w for w in sent_corrected.split() if len(w) > 1]
                '''
                cnt_cnd = long_word_processing(word, concept_lst) # list mapping for concept word
                lst.extend(cnt_cnd)
                usr_cnd = long_word_processing(word, usrword_lst)
                lst.extend(usr_cnd)
            else: 
                lst.append(word)

    #adverb_lst = list_reader(adv_path, False)
    #stopword_lst = list_reader(stw_path, False)
    
    lst = processor.stopword_remove(lst, stopword_lst)
    lst = processor.adverb_remove(lst, adverb_lst)

    return lst

def noun_chunk_model(text, stopword_lst=stopword_lst, char_num = 4):
    lst = []
    uni = processor.load_dict()

    line = processor.pre_processing(text)
    line = processor.unicode_normalize(line,uni)

    word_list = hanja_text_to_list(line)
    for word in word_list:
        word_length = len(word)
        word = word.replace('\n', '').replace(' ', '').replace(',','')
        
        is_all_hanja = True
        for char in word:
            if hangul.is_hangul(char): is_all_hanja = False
            
        is_num_word, is_skip, cnt = processor.num_word_check(word)
        if is_num_word or is_skip: continue
        is_date, is_year = processor.date_check(word, cnt)
        if is_date or is_year: continue
        
        word = processor.last_char_remove(word)
        word = processor.prefix_remove(word)
        
        if is_all_hanja and word_length > 1 and word.isalpha()== True: 
            if word_length > char_num: #긴 단어 처리
                sent_corrected, tags = model.correct(word, rules=rule_dict) # soyspacing 모델, space_rule을 적용
                candidate = [w for w in sent_corrected.split() if len(w) > 1]
                candidate = [processor.prefix_remove(cnd) for cnd in candidate]
                lst.extend(candidate)
            else: lst.append(word)
    
    #adverb_lst = list_reader(adv_path, False)
    #stopword_lst = list_reader(stw_path, False)

    lst = processor.stopword_remove(lst, stopword_lst)
    lst = processor.adverb_remove(lst, adverb_lst)

    return lst

def long_word_processing(word, concept_lst=concept_lst):
    candidate, rm_lst = [], []
    for concept in concept_lst:
        if concept in word: candidate.append(concept)
    candidate.sort(key=len)
    length = len(candidate)
    for i in range(0, length-1):
        for j in range(i+1, length):
            if candidate[j].startswith(candidate[i]):rm_lst.append(candidate[i])
            elif candidate[j].endswith(candidate[i]):rm_lst.append(candidate[i])
    for remove_word in set(rm_lst):
        candidate.remove(remove_word)
    return candidate

def hanja_text_to_list(line):
    is_all_hanja = True
    for char in line: 
        if hangul.is_hangul(char): is_all_hanja = False
    if is_all_hanja : #모두 한자, 띄어쓰기가 있는 경우: 띄어쓰기로 쪼갬
        all_hanja = [word for word in line.split() if len(word) > 1]
        hanja_list = all_hanja
    else:             #국한문혼용, 한글로 된 부분과 한자로 된 부분을 분리
        words_list = [word for word in split_hanja(line) if len(word) > 1]
        hanja_list = words_list
    return hanja_list



     
def word_count(word_list):
    wordCount = {}
    for word in word_list:
        wordCount[word] = wordCount.get(word, 0) + 1
        sorted_dict = sorted(wordCount.items(), key=lambda x: x[1], reverse=True)
    return sorted_dict

def co_occurence_count(word_lst):
    co_dic = {}
    for i,a in enumerate(word_lst):
        #for b in word_lst[i+1:i+11]:
        for b in word_lst[i+1:]:
            if a == b: continue
            #print(i,a,b)
            if a > b: c, d = b, a
            else: c, d = a, b
            co_dic[c, d] = co_dic.get((c, d), 0) + 1
    return co_dic

def extract_window(lst, kwd, window=5):
    wpairs = []    

    if kwd in lst:
        locs = [i for i, word in enumerate(lst) if word in kwd]
        last = len(lst)

        ext_pairs = []
        pairs = []; 
        for loc in locs:
            pair = (max(0,loc-window),min(loc+(window+1), last)) # before and after n words
            pairs.append(pair)

            for begin, end in sorted(pairs): # merge ranges overlap
                if ext_pairs and ext_pairs[-1][1] >= begin - 1: ext_pairs[-1][1] = max(ext_pairs[-1][1], end)
                else: ext_pairs.append([begin, end])

        for ext_pair in ext_pairs:
            p = lst[ext_pair[0]:ext_pair[1]]
            #rgs = [str(ext_pair[0]), str(ext_pair[1])]
            wpairs.append(p)

    return wpairs



# -*- coding: utf-8 -*-
import moko.noun_chunker as nc
import moko.term_analyzer as ta

def moko_demo():
    text = "泱泱大風이 固由於萬籟齊應이나 其初也엔 起於一蓬之末고 彼文明國之所謂 文明이 固謂其國民全軆之文明이나 其文明開發之原動力은"

    dct_lst = nc.noun_chunk_dict(text,char_num=4)
    print("=== noun list with a dictionary ===")
    print(dct_lst)

    mdl_lst = nc.noun_chunk_model(text)
    print("=== noun list with a model ===")
    print(mdl_lst)
    
    win = ta.extract_window(dct_lst,"文明",2)
    print(win)
    win = ta.extract_window(mdl_lst,"文明",2)
    print(win)

moko_demo()


from setuptools import setup, find_packages

with open('README.md', encoding='utf-8') as f:
    long_description = f.read()

setup(name='moko',
    version='0.1.0.15',
    description='Modern Korean NLP Package',
    long_description = long_description,
    long_description_content_type='text/markdown',
    author='yk.jeong, m.kim',
    author_email='yookyungjeong@gmail.com, munui0822@gmail.com',
    url='https://cmks.yonsei.ac.kr/cmks/database.htm?ch=2',
    license='MIT License',
    py_modules=['moko'],
    python_requires='>=3.6',
    install_requires=['soyspacing','hanja'],
    packages=['moko'],
    package_data={'moko':['data/*','model/*']}
)




from collections import defaultdict

import csv

def character_reader(character_file_path):
    character_list = []
    with open(character_file_path, 'r', encoding = 'utf-8') as character_file:
        while True:
            line = character_file.readline().replace('\n', '')
            if not line: break
            character_list.append(line)
        return character_list

def term_reader(term_extra):
    term_extra_dic = defaultdict(list)
    for line in csv_file_reader(term_extra):
        term_extra_dic[line[0]] = line[1]
    return term_extra_dic

def list_reader(file_path, header = True):
    filereader = csv_file_reader(file_path)
    if header:
        lst = [line[1] for line in filereader if len(line[1]) > 1] #line[1]:concept
    else:
        lst = [line for line in filereader]
        lst = sum(lst, [])
    return lst

def file_reader(file_path):
    f = open(file_path, 'r', encoding = 'utf-8')
    return f
    
def file_writer(file_path):
    f = open(file_path, 'w', encoding = 'utf-8')
    return f

def csv_file_reader(file_path):
    csv_file_reader = open(file_path, 'r', newline='', encoding = 'utf-8')
    filereader = csv.reader(csv_file_reader)
    next(filereader)
    return filereader

def csv_file_writer(file_path):
    f = open(file_path, 'w', newline='', encoding = 'utf-8')
    csvwriter = csv.writer(f)
    return csvwriter
